# Imports do próprio módulo

from inewave.config import MAX_ANOS_ESTUDO, MAX_ITERS, MAX_MESES_ESTUDO
from inewave.config import MAX_REES
from inewave.config import MESES_DF

# Imports de módulos externos
from cfinterface.components.block import Block
from cfinterface.components.line import Line
from cfinterface.components.field import Field
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField
from datetime import timedelta
import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from typing import IO, List, Optional


class BlocoEafPastTendenciaHidrolPMO(Block):
    """
    Bloco de informações de afluências passadas para
    tendência hidrológica localizado no arquivo `pmo.dat`.
    """

    BEGIN_PATTERN = "ENERGIAS AFLUENTES PASSADAS PARA A TENDENCIA HIDROLOGICA"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        ree_field: List[Field] = [LiteralField(10, 1)]
        ena_fields: List[Field] = [
            FloatField(11, 11 * (i + 1), 2) for i in range(len(MESES_DF))
        ]
        self.__line = Line(ree_field + ena_fields)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoEafPastTendenciaHidrolPMO):
            return False
        bloco: BlocoEafPastTendenciaHidrolPMO = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            df = pd.DataFrame(tabela, columns=MESES_DF)
            df["REE"] = rees
            df = df[["REE"] + MESES_DF]
            return df

        # Pula as linhas iniciais
        for _ in range(4):
            file.readline()

        # Variáveis auxiliares
        rees: List[str] = []
        tabela = np.zeros((MAX_REES, len(MESES_DF)))
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if "X-------" in linha:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            # Senão, processa os dados
            dados_linha: list = self.__line.read(linha)
            rees.append(dados_linha[0])
            tabela[i, :] = dados_linha[1:]
            i += 1


class BlocoEafPastCfugaMedioPMO(Block):
    """
    Bloco de informações de afluências passadas para
    tendência hidrológica localizado no arquivo `pmo.dat`.
    """

    BEGIN_PATTERN = (
        "ENERGIAS AFLUENTES PASSADAS EM REFERENCIA A PRIMEIRA CONFIG"
    )
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        ree_field: List[Field] = [LiteralField(10, 1)]
        ena_fields: List[Field] = [
            FloatField(11, 11 * (i + 1), 2) for i in range(len(MESES_DF))
        ]
        self.__line = Line(ree_field + ena_fields)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoEafPastCfugaMedioPMO):
            return False
        bloco: BlocoEafPastCfugaMedioPMO = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            df = pd.DataFrame(tabela, columns=MESES_DF)
            df["REE"] = rees
            df = df[["REE"] + MESES_DF]
            return df

        # Pula as linhas iniciais
        for _ in range(4):
            file.readline()

        # Variáveis auxiliares
        rees: List[str] = []
        tabela = np.zeros((MAX_REES, len(MESES_DF)))
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if "X-------" in linha:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            # Senão, processa os dados
            dados_linha: list = self.__line.read(linha)
            rees.append(dados_linha[0])
            tabela[i, :] = dados_linha[1:]
            i += 1


class BlocoConvergenciaPMO(Block):
    """
    Bloco com as informações de convergência do NEWAVE obtidas
    no arquivo `pmo.dat`.
    """

    BEGIN_PATTERN = "    ITER               LIM.INF.        "
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        # Cria a estrutura de uma linha da tabela
        iter_field: List[Field] = [IntegerField(4, 4)]
        conv_fields: List[Field] = [
            FloatField(23, 9 + 23 * i, 2) for i in range(6)
        ]
        self.__line = Line(iter_field + conv_fields)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoConvergenciaPMO):
            return False
        bloco: BlocoConvergenciaPMO = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df() -> pd.DataFrame:
            df = pd.DataFrame(tabela)
            df.columns = [
                "Iteração",
                "Lim. Inf. ZINF",
                "ZINF",
                "Lim. Sup. ZINF",
                "ZSUP",
                "Delta ZINF",
                "ZSUP Iteração",
            ]
            df = df.astype({"Iteração": "int64"})
            df["Tempo"] = tempos
            return df

        # Salta as duas linhas iniciais
        for _ in range(3):
            file.readline()
        # Variáveis auxiliares
        tabela = np.zeros((3 * MAX_ITERS, 7), dtype=np.float64)
        tempos: List[Optional[timedelta]] = []
        i = 0
        while True:
            linha: str = file.readline()
            # Confere se já acabou
            if len(linha) < 3:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            # Se não tem nada na coluna de iteração, ignora a linha
            if linha[4:9] == "     ":
                continue
            # Senão, lê a linha
            tabela[i, :] = self.__line.read(linha)
            # Lê o tempo, que tem formato personalizado
            if "min" in linha[153:168]:
                tempo = linha[153:168]
                h = int(tempo.split("h")[0])
                min = int(tempo.split("h")[1].split("min")[0])
                s = float(tempo.split("min")[1].split("s")[0])
                ts = timedelta(hours=h, minutes=min, seconds=s)
                tempos.append(ts)
            else:
                tempos.append(None)
            i += 1


class BlocoConfiguracoesExpansaoPMO(Block):
    """
    Bloco de informações sobre as configurações de expansão
    do sistema existentes no arquivo `pmo.dat`.
    """

    BEGIN_PATTERN = "CONFIGURACOES POR"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        # Cria a estrutura de uma linha da tabela
        self.__line = Line([IntegerField(6, 6 * (i + 1)) for i in range(13)])

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoConfiguracoesExpansaoPMO):
            return False
        bloco: BlocoConfiguracoesExpansaoPMO = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            df = pd.DataFrame(tabela, columns=["Ano"] + MESES_DF)
            return df

        # Pula as linhas iniciais
        for _ in range(6):
            file.readline()
        # Variáveis auxiliares
        tabela = np.zeros((MAX_ANOS_ESTUDO, len(MESES_DF) + 1), dtype=np.int64)
        i = 0
        while True:
            linha: str = file.readline()
            # Confere se acabou
            if len(linha) < 3:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            # Lê mais uma linha
            tabela[i, :] = self.__line.read(linha)
            i += 1


class BlocoMARSPMO(Block):
    """
    Bloco de informações sobre o modelo MARS ajustado
    para as retas de perdas por engolimento máximo
    existentes no arquivo `pmo.dat`.
    """

    BEGIN_PATTERN = "PARAMETROS DAS RETAS DE PERDAS POR ENGOLIMENTO MAXIMO"
    END_PATTERN = 'ENERGIA FIO D"AGUA LIQUIDA|CEPEL'

    MAX_RETAS_MARS = 3

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        # Cria a estrutura de uma linha da tabela
        self.__line = Line(
            [
                IntegerField(12, 1),
                FloatField(12, 14, 5, format="E"),
                FloatField(12, 27, 5, format="E"),
            ]
        )
        self.__ree_field = LiteralField(10, 6)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoMARSPMO):
            return False
        bloco: BlocoMARSPMO = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            colunas = ["Reta", "Coeficiente Angular", "Constante"]
            df = pd.DataFrame(tabela, columns=colunas)
            df["REE"] = rees
            df = df[["REE"] + colunas]
            df = df.astype({"Reta": "int64"})
            return df

        # Variáveis auxiliares
        ree_atual = ""
        rees: List[str] = []
        tabela = np.zeros(
            (BlocoMARSPMO.MAX_RETAS_MARS * MAX_REES * MAX_MESES_ESTUDO, 3),
            dtype=np.float64,
        )
        i = 0
        while True:
            linha: str = file.readline()
            if self.ends(linha):
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            elif "REE:" in linha:
                ree_atual = self.__ree_field.read(linha)
            else:
                dados_linha = self.__line.read(linha)
                if dados_linha[0] is not None:
                    rees.append(ree_atual)
                    tabela[i, :] = dados_linha
                    i += 1


class BlocoRiscoDeficitENSPMO(Block):
    """
    Bloco de informações sobre os riscos de déficit e
    ENS (energia não suprida) existentes no arquivo `pmo.dat`.
    """

    BEGIN_PATTERN = "RISCO ANUAL DE DEFICIT E E\(ENS\) \(%\)"  # noqa
    END_PATTERN = ""

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoRiscoDeficitENSPMO):
            return False
        bloco: BlocoRiscoDeficitENSPMO = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df() -> pd.DataFrame:
            df = pd.DataFrame(tabela)
            cols = ["Ano"]
            for sub in subsistemas:
                cols += [f"Risco - {sub}", f"EENS - {sub}"]
            df.columns = cols
            df = df.astype({"Ano": "int64"})
            return df

        # Pula as três linhas iniciais
        for _ in range(3):
            file.readline()
        # Identifica os subsistemas em questão e constroi
        # a estrutura da linha da tabela a ser lida
        subsistemas = [
            s.strip() for s in file.readline().split(" ") if len(s) > 1
        ]
        # Pula três linhas adicionais
        for _ in range(3):
            file.readline()
        campos: List[Field] = [IntegerField(4, 1)]
        ultima_coluna = 6
        for _ in subsistemas:
            campos.append(FloatField(6, ultima_coluna, 2))
            ultima_coluna += 6
            campos.append(FloatField(8, ultima_coluna, 2))
            ultima_coluna += 8
        self.__line = Line(campos)
        # Variáveis auxiliares
        tabela = np.zeros((MAX_ANOS_ESTUDO, 2 * len(subsistemas) + 1))
        i = 0
        while True:
            linha: str = file.readline()
            # Confere se acabou
            if len(linha) < 3:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            tabela[i, :] = self.__line.read(linha)
            i += 1


class BlocoCustoOperacaoPMO(Block):
    """
    Bloco de informações sobre os custos de operação categorizados
    existentes no arquivo `pmo.dat`.
    """

    BEGIN_PATTERN = "PARCELA           V.ESPERADO"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        # Cria a estrutura de uma linha da tabela
        self.__line = Line(
            [
                LiteralField(18, 13),
                FloatField(13, 32, 2),
                FloatField(13, 46, 2),
                FloatField(7, 60, 2),
            ]
        )

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoCustoOperacaoPMO):
            return False
        bloco: BlocoCustoOperacaoPMO = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df() -> pd.DataFrame:
            cols = ["Valor Esperado", "Desvio Padrão do VE", "(%)"]
            df = pd.DataFrame(tabela, columns=cols)
            df["Parcela"] = parcelas
            df = df[["Parcela"] + cols]
            return df

        # Salta duas linhas
        for _ in range(2):
            file.readline()
        parcelas: List[str] = []
        tabela = np.zeros((30, 3), dtype=np.float64)
        i = 0
        while True:
            linha = file.readline()
            if "----------------" in linha:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            dados_linha = self.__line.read(linha)
            parcelas.append(dados_linha[0])
            tabela[i, :] = dados_linha[1:]
            i += 1


class BlocoCustoOperacaoTotalPMO(Block):
    """
    Bloco de informações sobre os custos de operação categorizados
    existentes no arquivo `pmo.dat`.
    """

    BEGIN_PATTERN = "           VALOR ESPERADO TOTAL:"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        # Cria a estrutura de uma linha da tabela
        self.__line = Line([FloatField(24, 50, 2)])

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoCustoOperacaoTotalPMO):
            return False
        bloco: BlocoCustoOperacaoTotalPMO = o
        if not all(
            [
                isinstance(self.data, list),
                isinstance(o.data, list),
            ]
        ):
            return False
        else:
            return all([sd == d for sd, d in zip(self.data, bloco.data)])

    # Override
    def read(self, file: IO):
        data = [0, 0]
        for i in range(2):
            data[i] = self.__line.read(file.readline())[0]
        self.data = data



from inewave.newave.modelos.vazoes import RegistroVazoesPostos
from inewave.newave.vazoes import Vazoes

from os.path import join
from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch


ARQ_TEST = "./tests/mocks/arquivos"


def test_registro_vazoesposto_vazoes():

    r = RegistroVazoesPostos()
    with open(join(ARQ_TEST, "vazoes.dat"), "rb") as fp:
        r.read(fp, storage="BINARY")

    assert len(r.data) == 320


def test_atributos_encontrados_vazoes():
    h = Vazoes.le_arquivo(ARQ_TEST)
    assert h.vazoes is not None


def test_atributos_nao_encontrados_vazoes():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        ad = Vazoes.le_arquivo("")
        assert ad.vazoes is None


def test_eq_vazoes():
    h1 = Vazoes.le_arquivo(ARQ_TEST)
    h2 = Vazoes.le_arquivo(ARQ_TEST)
    assert h1 == h2


def test_neq_vazoes():
    h1 = Vazoes.le_arquivo(ARQ_TEST)
    h2 = Vazoes.le_arquivo(ARQ_TEST)
    h2.vazoes.iloc[0, 0] = -1
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        h2.escreve_arquivo("")
        assert h1 != h2


def test_leitura_escrita_vazoes():
    h1 = Vazoes.le_arquivo(ARQ_TEST)
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        h1.escreve_arquivo("", "")
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(1, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data=b"".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        h2 = Vazoes.le_arquivo("")
        assert h1 == h2


def test_leitura_escrita_editando_vazoes():
    h1 = Vazoes.le_arquivo(ARQ_TEST)
    vaz = h1.vazoes
    num_vazoes_original = vaz.shape[0]
    h1.vazoes.loc[vaz.shape[0]] = 0
    m_escrita: MagicMock = mock_open(read_data="")
    # Testa aumentando a quantidade de vazões
    with patch("builtins.open", m_escrita):
        h1.escreve_arquivo("", "")
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(1, len(chamadas) - 1)
        ]
        assert len(linhas_escritas) == num_vazoes_original
    # Testa reduzindo a quantidade de vazões
    num_vazoes_reduzidas = 10
    h1.vazoes.drop(
        index=list(range(num_vazoes_reduzidas, num_vazoes_original + 1)),
        inplace=True,
    )
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        h1.escreve_arquivo("", "")
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(1, len(chamadas) - 1)
        ]
        assert len(linhas_escritas) == num_vazoes_reduzidas



from cfinterface.files.registerfile import RegisterFile
from inewave.newave.modelos.vazoes import RegistroVazoesPostos
import pandas as pd  # type: ignore


from typing import TypeVar, List, Optional


class Vazoes(RegisterFile):
    """
    Armazena os dados de entrada do NEWAVE referentes ao cadastro das
    usinas hidroelétricas.
    """

    T = TypeVar("T")

    REGISTERS = [RegistroVazoesPostos]
    POSTOS = 320
    STORAGE = "BINARY"

    def __init__(self, data=...) -> None:
        super().__init__(data)
        self.__df: Optional[pd.DataFrame] = None
        RegistroVazoesPostos.set_postos(self.POSTOS)

    @classmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="vazoes.dat") -> "Vazoes":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="vazoes.dat"):
        self.__atualiza_registros()
        self.write(diretorio, nome_arquivo)

    def __monta_df_de_registros(self) -> Optional[pd.DataFrame]:
        registros: List[RegistroVazoesPostos] = [
            r for r in self.data.of_type(RegistroVazoesPostos)
        ]
        if len(registros) == 0:
            return None
        df = pd.DataFrame(columns=list(range(1, self.__class__.POSTOS + 1)))
        for i, r in enumerate(registros):
            df.loc[i] = r.data

        df = df.astype({i: int for i in range(1, self.__class__.POSTOS + 1)})
        return df

    def __atualiza_registros(self):
        registros: List[RegistroVazoesPostos] = [r for r in self.data][1:]
        n_registros = len(registros)
        n_meses = self.vazoes.shape[0]
        # Deleta os registros que sobraram
        for i in range(n_meses, n_registros):
            self.data.remove(registros[i])
        # Cria registros se faltaram
        for i in range(n_registros, n_meses):
            self.data.append(RegistroVazoesPostos())
        # Atualiza os dados
        for (_, linha), r in zip(self.vazoes.iterrows(), registros):
            r.data = linha.tolist()

    @property
    def vazoes(self) -> pd.DataFrame:
        """
        Obtém a tabela com os dados de vazão existentes no arquivo
        binário.

        - 1...N (`int`) - onde N é o número de postos

        :return: A tabela com as vazões por posto
        :rtype: pd.DataFrame
        """
        if self.__df is None:
            self.__df = self.__monta_df_de_registros()
        return self.__df

    @vazoes.setter
    def vazoes(self, df: pd.DataFrame):
        self.__df = df



from inewave.newave.modelos.hidr import RegistroUHEHidr
from inewave.newave.hidr import Hidr

from os.path import join
from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch


ARQ_TEST = "./tests/mocks/arquivos"


def test_registro_uhe_hidr():

    r = RegistroUHEHidr()
    with open(join(ARQ_TEST, "hidr.dat"), "rb") as fp:
        r.read(fp, storage="BINARY")

    assert len(r.data) == 111
    assert r.nome == "CAMARGOS"


def test_atributos_encontrados_hidr():
    h = Hidr.le_arquivo(ARQ_TEST)
    assert h.cadastro is not None


def test_atributos_nao_encontrados_hidr():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        ad = Hidr.le_arquivo("")
        assert ad.cadastro is None


def test_eq_hidr():
    h1 = Hidr.le_arquivo(ARQ_TEST)
    h2 = Hidr.le_arquivo(ARQ_TEST)
    assert h1 == h2


def test_neq_hidr():
    h1 = Hidr.le_arquivo(ARQ_TEST)
    h2 = Hidr.le_arquivo(ARQ_TEST)
    h2.cadastro.iloc[0, 0] = "TESTE"
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        h2.escreve_arquivo("")
        assert h1 != h2


def test_leitura_escrita_hidr():
    h1 = Hidr.le_arquivo(ARQ_TEST)
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        h1.escreve_arquivo("", "")
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(1, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data=b"".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        h2 = Hidr.le_arquivo("")
        assert h1 == h2



from typing import Optional
from cfinterface.components.register import Register
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField


class RegistroEolicaSubmercado(Register):
    """
    Registro que contém uma relação entre usina eólica e submercado.
    """

    IDENTIFIER = "EOLICA-SUBMERCADO"
    IDENTIFIER_DIGITS = 17
    LINE = Line(
        [
            IntegerField(),
            IntegerField(),
        ],
        delimiter=";",
    )

    @property
    def codigo_eolica(self) -> Optional[int]:
        return self.data[0]

    @codigo_eolica.setter
    def codigo_eolica(self, c: int):
        self.data[0] = c

    @property
    def codigo_submercado(self) -> Optional[int]:
        return self.data[1]

    @codigo_submercado.setter
    def codigo_submercado(self, c: int):
        self.data[1] = c



# Rotinas de testes associadas ao arquivo eolica-cadastro.csv do NEWAVE
from datetime import datetime
from inewave.newave.modelos.eolicacadastro import (
    RegistroEolicaCadastro,
    RegistroEolicaCadastroConjuntoAerogeradores,
    RegistroEolicaCadastroAerogerador,
    RegistroEolicaConjuntoAerogeradoresQuantidadeOperandoPeriodo,
    RegistroEolicaConjuntoAerogeradoresPotenciaEfetiva,
)

from inewave.newave import EolicaCadastro

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.eolicacadastro import (
    MockRegistroEolicaCadastro,
    MockRegistroEolicaCadastroAerogerador,
    MockRegistroEolicaCadastroConjuntoAerogeradores,
    MockRegistroEolicaConjuntoAerogeradoresQuantidadeOperandoPeriodo,
    MockRegistroEolicaConjuntoAerogeradoresPotenciaEfetiva,
    MockEolicaCadastro,
)


def test_registro_eolica_cadastro_eolicacadastro():

    m: MagicMock = mock_open(read_data="".join(MockRegistroEolicaCadastro))
    r = RegistroEolicaCadastro()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, "NEInterior", "", 1]
    assert r.codigo_eolica == 1
    r.codigo_eolica = 5
    assert r.nome_eolica == "NEInterior"
    r.nome_eolica = "Teste"
    assert r.identificador_eolica == ""
    r.identificador_eolica = "   "
    assert r.quantidade_conjuntos == 1
    r.quantidade_conjuntos = 5


def test_registro_cadastro_aerogerador_eolicacadastro():

    m: MagicMock = mock_open(
        read_data="".join(MockRegistroEolicaCadastroAerogerador)
    )
    r = RegistroEolicaCadastroAerogerador()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, 1, 0, 0, 0, 0, 6058.890, 0, 0]
    assert r.codigo_eolica == 1
    r.codigo_eolica = 2
    assert r.indice_conjunto == 1
    r.indice_conjunto = 1
    assert r.velocidade_cutin == 0
    r.velocidade_cutin = 5
    assert r.velocidade_nominal == 0
    r.velocidade_nominal = 3
    assert r.velocidade_cutout == 0
    r.velocidade_cutout = 10
    assert r.potencia_velocidade_cutin == 0
    r.potencia_velocidade_cutin = 56
    assert r.potencia_velocidade_nominal == 6058.890
    r.potencia_velocidade_nominal = 600
    assert r.potencia_velocidade_cutout == 0
    r.potencia_velocidade_cutout = 10000
    assert r.altura_torre == 0
    r.altura_torre = 100


def test_registro_cadastro_conjunto_eolicacadastro():

    m: MagicMock = mock_open(
        read_data="".join(MockRegistroEolicaCadastroConjuntoAerogeradores)
    )
    r = RegistroEolicaCadastroConjuntoAerogeradores()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, 1, "NEInterior_cj", 1]
    assert r.codigo_eolica == 1
    r.codigo_eolica = 2
    assert r.indice_conjunto == 1
    r.indice_conjunto = 1
    assert r.nome_conjunto == "NEInterior_cj"
    r.nome_conjunto = "Teste"
    assert r.quantidade_aerogeradores == 1
    r.quantidade_aerogeradores = 2


def test_registro_cadastro_operacao_eolicacadastro():

    m: MagicMock = mock_open(
        read_data="".join(
            MockRegistroEolicaConjuntoAerogeradoresQuantidadeOperandoPeriodo
        )
    )
    r = RegistroEolicaConjuntoAerogeradoresQuantidadeOperandoPeriodo()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, 1, datetime(2021, 1, 1), datetime(2030, 12, 1), 1]
    assert r.codigo_eolica == 1
    r.codigo_eolica = 2
    assert r.indice_conjunto == 1
    r.indice_conjunto = 1
    assert r.periodo_inicial == datetime(2021, 1, 1)
    r.periodo_inicial = datetime(2024, 1, 1)
    assert r.periodo_final == datetime(2030, 12, 1)
    r.periodo_final = datetime(2025, 12, 1)
    assert r.numero_aerogeradores == 1
    r.numero_aerogeradores = 3


def test_registro_potenciaefetiva_eolicacadastro():

    m: MagicMock = mock_open(
        read_data="".join(
            MockRegistroEolicaConjuntoAerogeradoresPotenciaEfetiva
        )
    )
    r = RegistroEolicaConjuntoAerogeradoresPotenciaEfetiva()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [
        1,
        1,
        datetime(2021, 1, 1),
        datetime(2030, 12, 1),
        6058.890,
    ]
    assert r.codigo_eolica == 1
    r.codigo_eolica = 2
    assert r.indice_conjunto == 1
    r.indice_conjunto = 1
    assert r.periodo_inicial == datetime(2021, 1, 1)
    r.periodo_inicial = datetime(2024, 1, 1)
    assert r.periodo_final == datetime(2030, 12, 1)
    r.periodo_final = datetime(2025, 12, 1)
    assert r.potencia_efetiva == 6058.890
    r.potencia_efetiva = 5


def test_atributos_encontrados_eolicacadastro():
    m: MagicMock = mock_open(read_data="".join(MockEolicaCadastro))
    with patch("builtins.open", m):
        e = EolicaCadastro.le_arquivo("")
        assert len(e.eolica_cadastro()) > 0
        assert len(e.eolica_cadastro_aerogerador()) > 0
        assert len(e.eolica_cadastro_conjunto_aerogeradores()) > 0
        assert (
            len(e.eolica_conjunto_aerogeradores_quantidade_operando_periodo())
            > 0
        )
        assert (
            len(e.eolica_conjunto_aerogeradores_potencia_efetiva_periodo()) > 0
        )


def test_eq_eolicacadastro():
    m: MagicMock = mock_open(read_data="".join(MockEolicaCadastro))
    with patch("builtins.open", m):
        cf1 = EolicaCadastro.le_arquivo("")
        cf2 = EolicaCadastro.le_arquivo("")
        assert cf1 == cf2


def test_neq_eolicacadastro():
    m: MagicMock = mock_open(read_data="".join(MockEolicaCadastro))
    with patch("builtins.open", m):
        cf1 = EolicaCadastro.le_arquivo("")
        cf2 = EolicaCadastro.le_arquivo("")
        cf2.deleta_registro(cf1.eolica_cadastro()[0])
        assert cf1 != cf2


def test_leitura_escrita_eolicacadastro():
    m_leitura: MagicMock = mock_open(read_data="".join(MockEolicaCadastro))
    with patch("builtins.open", m_leitura):
        cf1 = EolicaCadastro.le_arquivo("")
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        cf1.escreve_arquivo("", "")
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(2, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        cf2 = EolicaCadastro.le_arquivo("")
        assert cf1 == cf2



from inewave.nwlistop.modelos.blocos.ree import REE
from inewave.nwlistop.modelos.arquivos.arquivoreepatamar import (
    ArquivoREEPatamar,
)
from inewave.nwlistop.modelos.ghmaxr import GHAnos


class Ghmaxr(ArquivoREEPatamar):
    """
    Armazena os dados das saídas referentes à geração hidraulica máxima
    considerando restrições elétricas por patamar, por REE.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `ghmax00x.out`, onde x varia conforme o
    REE em questão.

    """

    BLOCKS = [
        REE,
        GHAnos,
    ]

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="ghmaxr001.out"
    ) -> "Ghmaxr":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="ghmaxr001.out"):
        self.write(diretorio, nome_arquivo)



import os
from abc import abstractmethod
from traceback import print_exc
import pandas as pd  # type: ignore


class LeituraCSV:
    """
    Classe com utilidades gerais para leitura de arquivos
    .csv do NEWAVE.
    """

    def __init__(self, diretorio: str):
        self._diretorio = diretorio
        self._dados = pd.DataFrame()

    def _le_tabela_arquivo(self, caminho: str):
        """
        Faz a leitura dos blocos de dados do arquivo.
        """
        self._dados = pd.read_csv(
            caminho, sep=",", encoding="utf-8", index_col=1
        )

    def _le_arquivo_em_diretorio(
        self, diretorio: str, nome_arquivo: str
    ) -> None:
        """
        Faz a leitura do arquivo em um diretorio.
        """
        try:
            caminho = os.path.join(diretorio, nome_arquivo)
            self._le_tabela_arquivo(caminho)
            self.processa_dados_lidos()
        except Exception:
            print_exc()

    def le_arquivo(self, nome_arquivo: str) -> pd.DataFrame:
        """
        Método para ler um arquivo e retornar o objeto
        devido da classe em particular.
        """
        self._le_arquivo_em_diretorio(self._diretorio, nome_arquivo)
        return self._dados

    @abstractmethod
    def processa_dados_lidos(self):
        """ """
        pass

    @property
    def dados(self) -> pd.DataFrame:
        """
        Dados de leitura obtidos pelo blocos, organizados
        para realizar a criação do objeto associado.
        """
        return self._dados



MockBlocoConfUTE = [
    " NUM  NOME           SSIS  U.EXIS CLASSE\n",
    " XXXX XXXXXXXXXXXX   XXXX     XX   XXXX\n",
    "    1 ANGRA 1           1     EX      1\n",
    "   13 ANGRA 2           1     EX     13\n",
    "  211 BAIXADA FLU       1     EX    211\n",
    "   97 CUBATAO           1     EE     97\n",
    "   12 CUIABA G CC       1     EE     12\n",
    "  153 DAIA              1     EE    153\n",
    "   65 DO ATLAN_CSA      1     EE     65\n",
    "  183 DO ATLANTICO      1     EE    183\n",
    "  137 GNA I             1     NE    137\n",
    "  238 GNA P. ACU 3      1     NE    238\n",
    "  155 GOIANIA II        1     EE    155\n",
    "   63 IBIRITE           1     EX     63\n",
    "    2 IGARAPE           1     EE      2\n",
    "   54 JUIZ DE FORA      1     EE     54\n",
    "   15 LINHARES          1     EE     15\n",
    "  236 MARLIM AZUL       1     NE    236\n",
    "  110 N.PIRATINING      1     EE    110\n",
    "  171 NORTEFLU-1        1     EX    171\n",
    "  172 NORTEFLU-2        1     EX    172\n",
    "  173 NORTEFLU-3        1     EE    173\n",
    "  174 NORTEFLU-4        1     EE    174\n",
    "  229 ONCA PINTADA      1     NE    229\n",
    "  182 PALMEIRAS GO      1     EE    182\n",
    "   50 PIRAT.12 G        1     EE     50\n",
    "  230 PREDILECTA        1     NE    230\n",
    "    9 R.SILVEIRA        1     EE      9\n",
    "   62 SEROPEDICA        1     EE     62\n",
    "    4 ST.CRUZ 34        1     EE      4\n",
    "   86 ST.CRUZ NOVA      1     EE     86\n",
    "  146 STA VITORIA       1     NE    146\n",
    "  194 T.NORTE 2         1     EE    194\n",
    "  223 TERMOIRAPE I      1     NE    223\n",
    "   90 TERMOMACAE        1     EE     90\n",
    "   47 TERMORIO          1     EE     47\n",
    "   68 TRES LAGOAS       1     EE     68\n",
    "   49 VIANA             1     EE     49\n",
    "  108 XAVANTES          1     EE    108\n",
    "   48 ARAUCARIA         2     EE     48\n",
    "   37 ARGENTINA 1       2     EX     37\n",
    "   71 ARGENTINA 1B      2     EX     71\n",
    "   38 ARGENTINA 2A      2     EX     38\n",
    "   39 ARGENTINA 2B      2     EX     39\n",
    "   40 ARGENTINA 2C      2     EX     40\n",
    "   59 ARGENTINA 2D      2     EX     59\n",
    "  237 CAMBARA           2     NE    237\n",
    "  156 CANDIOTA 3        2     EE    156\n",
    "   64 CANOAS            2     EE     64\n",
    "  169 CISFRAMA          2     EX    169\n",
    "   28 FIGUEIRA          2     EE     28\n",
    "   26 J.LACERDA A1      2     EE     26\n",
    "   27 J.LACERDA A2      2     EE     27\n",
    "   25 J.LACERDA B       2     EE     25\n",
    "   24 J.LACERDA C       2     EE     24\n",
    "  107 PAMPA SUL         2     EE    107\n",
    "  149 SAO SEPE          2     EX    149\n",
    "   35 URUGUAIANA        2     EE     35\n",
    "  109 ALTOS             3     EE    109\n",
    "  111 ARACATI           3     EE    111\n",
    "  112 BAHIA I           3     EE    112\n",
    "  113 BATURITE          3     EE    113\n",
    "  235 CAMACARI MII      3     NE    235\n",
    "  166 CAMACARI PI       3     EX    166\n",
    "   52 CAMPINA GDE       3     EE     52\n",
    "  117 CAMPO MAIOR       3     EE    117\n",
    "  119 CAUCAIA           3     EE    119\n",
    "  121 CRATO             3     EE    121\n",
    "  125 ENGUIA PECEM      3     EE    125\n",
    "   42 FORTALEZA         3     EE     42\n",
    "   53 GLOBAL I          3     EE     53\n",
    "   55 GLOBAL II         3     EE     55\n",
    "  127 IGUATU            3     EE    127\n",
    "  133 JUAZEIRO N        3     EE    133\n",
    "   57 MARACANAU I       3     EE     57\n",
    "  135 MARAMBAIA         3     EE    135\n",
    "  164 MURICY            3     EE    164\n",
    "  138 NAZARIA           3     EE    138\n",
    "  167 P. PECEM I        3     EE    167\n",
    "  163 P. PECEM II       3     EX    163\n",
    "  224 P. SERGIPE I      3     NE    224\n",
    "  160 PAU FERRO I       3     EE    160\n",
    "  162 PECEM II          3     NE    162\n",
    "   98 PERNAMBU_III      3     EE     98\n",
    "  144 PETROLINA         3     EE    144\n",
    "  151 POTIGUAR          3     EE    151\n",
    "  161 POTIGUAR III      3     EE    161\n",
    "  147 PROSPERIDADE      3     EE    147\n",
    "  170 SUAPE II          3     EE    170\n",
    "  221 SYKUE I           3     EX    221\n",
    "   43 TERMOBAHIA        3     EE     43\n",
    "  152 TERMOCABO         3     EE    152\n",
    "   74 TERMOCAMACAR      3     EE     74\n",
    "   58 TERMOCEARA        3     EE     58\n",
    "  159 TERMOMANAUS       3     EE    159\n",
    "   67 TERMONE           3     EE     67\n",
    "   69 TERMOPB           3     EE     69\n",
    "   96 TERMOPE           3     EE     96\n",
    "   83 VALE DO ACU       3     EE     83\n",
    "  201 APARECIDA         4     EE    201\n",
    "  203 C. ROCHA          4     EE    203\n",
    "  318 FLORES LT1        4     EE    318\n",
    "  319 FLORES LT2        4     EE    319\n",
    "   73 GERAMAR I         4     EE     73\n",
    "   70 GERAMAR II        4     EE     70\n",
    "  317 IRANDUBA          4     EE    317\n",
    "  204 JARAQUI           4     EE    204\n",
    "  205 MANAUARA          4     EE    205\n",
    "  212 MARANHAO III      4     EE    212\n",
    "  436 MARAN IV L22      4     NE    436\n",
    "  437 MARAN IV L7       4     NE    437\n",
    "   36 MARANHAO IV       4     EE     36\n",
    "  421 MARAN V L22       4     NE    421   \n",
    "  422 MARAN V L7        4     NE    422   \n",
    "   21 MARANHAO V        4     EE     21\n",
    "  140 MAUA 3            4     EE    140\n",
    "  206 MAUA B3           4     EE    206\n",
    "  141 MAUA B4           4     EE    141\n",
    "  446 N.VEN 2 L22       4     NE    446 \n",
    "  447 N.VEN 2 L7        4     NE    447\n",
    "   46 N.VENECIA 2       4     EE     46\n",
    "  228 O. CANOAS 1       4     NE    228\n",
    "  116 PARNAIBA IV       4     EE    116\n",
    "  239 PARNAIBA V        4     NE    239\n",
    "  207 PONTA NEGRA       4     EE    207\n",
    "  176 PORTO ITAQUI      4     EE    176\n",
    "  209 TAMBAQUI          4     EE    209\n",
    "",
]



# Rotinas de testes associadas ao arquivo modif.dat do NEWAVE
from inewave.newave.modelos.modif import (
    USINA,
    VOLMIN,
    VOLMAX,
    NUMCNJ,
    NUMMAQ,
    VAZMIN,
    CFUGA,
    CMONT,
    VMAXT,
    VMINT,
    VMINP,
    VAZMINT,
)

from inewave.newave import Modif


from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.modif import (
    MockModif,
    MockUSINA,
    MockVOLMIN,
    MockVOLMAX,
    MockNUMCNJ,
    MockNUMMAQ,
    MockVAZMIN,
    MockCFUGA,
    MockCMONT,
    MockVMAXT,
    MockVMINT,
    MockVMINP,
    MockVAZMINT,
)


def test_registro_usina_modif():

    m: MagicMock = mock_open(read_data="".join(MockUSINA))
    r = USINA()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, "CAMARGOS"]
    assert r.codigo == 1
    assert r.nome == "CAMARGOS"


def test_registro_vazmin_modif():

    m: MagicMock = mock_open(read_data="".join(MockVAZMIN))
    r = VAZMIN()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [34]
    assert r.vazao == 34


def test_registro_vmaxt_modif():

    m: MagicMock = mock_open(read_data="".join(MockVMAXT))
    r = VMAXT()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [12, 2021, 61.310, "'%'"]
    assert r.mes == 12
    assert r.ano == 2021
    assert r.volume == 61.310
    assert r.unidade == "'%'"


def test_registro_vazmint_modif():

    m: MagicMock = mock_open(read_data="".join(MockVAZMINT))
    r = VAZMINT()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [10, 2021, 10.00]
    assert r.mes == 10
    assert r.ano == 2021
    assert r.vazao == 10.00


# def test_registro_volmin_modif():

#     m: MagicMock = mock_open(read_data="".join(MockVOLMIN))
#     r = VOLMIN()
#     with patch("builtins.open", m):
#         with open("", "") as fp:
#             r.read(fp)

#     assert r.data == [15563.63, "'h'"]
#     assert r.volume == 15563.63


# def test_registro_volmax_modif():

#     m: MagicMock = mock_open(read_data="".join(MockVOLMAX))
#     r = VOLMAX()
#     with patch("builtins.open", m):
#         with open("", "") as fp:
#             r.read(fp)

#     assert r.data == [15563.63, "'h'"]
#     assert r.volume == 15563.63
#     assert r.unidade == "'h'"


def test_registro_numcnj_modif():

    m: MagicMock = mock_open(read_data="".join(MockNUMCNJ))
    r = NUMCNJ()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [2]
    assert r.numero == 2


def test_registro_nummaq_modif():

    m: MagicMock = mock_open(read_data="".join(MockNUMMAQ))
    r = NUMMAQ()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [2, 5]
    assert r.conjunto == 2
    assert r.numero_maquinas == 5


def test_registro_vmint_modif():

    m: MagicMock = mock_open(read_data="".join(MockVMINT))
    r = VMINT()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [10, 2021, 20.0, "'%'"]
    assert r.mes == 10
    assert r.ano == 2021
    assert r.volume == 20.0
    assert r.unidade == "'%'"


def test_registro_vminp_modif():

    m: MagicMock = mock_open(read_data="".join(MockVMINP))
    r = VMINP()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [10, 2021, 20.0, "'%'"]
    assert r.mes == 10
    assert r.ano == 2021
    assert r.volume == 20.0
    assert r.unidade == "'%'"


def test_registro_cfuga_modif():

    m: MagicMock = mock_open(read_data="".join(MockCFUGA))
    r = CFUGA()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [10, 2021, 5.60]
    assert r.mes == 10
    assert r.ano == 2021
    assert r.nivel == 5.60


def test_registro_cmont_modif():

    m: MagicMock = mock_open(read_data="".join(MockCMONT))
    r = CMONT()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [10, 2021, 71.30]
    assert r.mes == 10
    assert r.ano == 2021
    assert r.nivel == 71.30


def test_atributos_encontrados_modif():
    m: MagicMock = mock_open(read_data="".join(MockModif))
    with patch("builtins.open", m):
        ad = Modif.le_arquivo("")
        assert len(ad.usina()) > 0
        assert len(ad.vazmin()) > 0
        assert len(ad.vmaxt()) > 0
        assert len(ad.vazmint()) > 0
        assert isinstance(ad.volmin(), VOLMIN)
        assert isinstance(ad.volmax(), VOLMAX)
        assert len(ad.numcnj()) > 0
        assert len(ad.nummaq()) > 0
        assert len(ad.vmint()) > 0
        assert len(ad.vminp()) > 0
        assert len(ad.cfuga()) > 0
        assert len(ad.cmont()) > 0


# def test_eq_patamar():
#     m: MagicMock = mock_open(read_data="".join(MockPatamar))
#     with patch("builtins.open", m):
#         cf1 = Patamar.le_arquivo("")
#         cf2 = Patamar.le_arquivo("")
#         assert cf1 == cf2


# def test_neq_patamar():
#     m: MagicMock = mock_open(read_data="".join(MockPatamar))
#     with patch("builtins.open", m):
#         cf1 = Patamar.le_arquivo("")
#         cf2 = Patamar.le_arquivo("")
#         cf2.numero_patamares = 0
#         assert cf1 != cf2


# def test_leitura_escrita_patamar():
#     m_leitura: MagicMock = mock_open(read_data="".join(MockPatamar))
#     with patch("builtins.open", m_leitura):
#         cf1 = Patamar.le_arquivo("")
#     m_escrita: MagicMock = mock_open(read_data="")
#     with patch("builtins.open", m_escrita):
#         cf1.escreve_arquivo("", "")
#         # Recupera o que foi escrito
#         chamadas = m_escrita.mock_calls
#         linhas_escritas = [
#             chamadas[i].args[0] for i in range(2, len(chamadas) - 1)
#         ]
#     m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
#     with patch("builtins.open", m_releitura):
#         cf2 = Patamar.le_arquivo("")
#         assert cf1 == cf2



from inewave.config import MAX_ANOS_ESTUDO, MAX_SUBMERCADOS, MESES_DF

from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.field import Field
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.floatfield import FloatField
from typing import List, IO, Optional
import pandas as pd  # type: ignore
import numpy as np  # type: ignore


class BlocoNumeroPatamaresDeficit(Section):
    """
    Bloco com o número de patamares de déficit considerados.
    """

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line(
            [
                IntegerField(3, 1),
            ]
        )
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoNumeroPatamaresDeficit):
            return False
        bloco: BlocoNumeroPatamaresDeficit = o
        if not all(
            [
                isinstance(self.data, int),
                isinstance(o.data, int),
            ]
        ):
            return False
        else:
            return self.data == bloco.data

    # Override
    def read(self, file: IO):
        for _ in range(3):
            self.__cabecalhos.append(file.readline())

        self.data = self.__linha.read(file.readline())[0]

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, int):
            raise ValueError(
                "Dados do sistema.dat não foram lidos com sucesso"
            )
        file.write(self.__linha.write([self.data]))


class BlocoCustosDeficit(Section):
    """
    Bloco com informações sobre o custo de déficit por
    patamar de déficit e o número de patamares de déficit.
    """

    FIM_BLOCO = " 999"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        campos_iniciais: List[Field] = [
            IntegerField(3, 1),
            LiteralField(12, 5),
            IntegerField(1, 17),
        ]
        campos_custos: List[Field] = [
            FloatField(7, 19 + i * 8, 2) for i in range(4)
        ]
        campos_cortes: List[Field] = [
            FloatField(5, 51 + i * 6, 3) for i in range(4)
        ]
        self.__linha = Line(campos_iniciais + campos_custos + campos_cortes)
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoCustosDeficit):
            return False
        bloco: BlocoCustosDeficit = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            cols = [f"Custo Pat. {p}" for p in range(1, 5)] + [
                f"Corte Pat. {p}" for p in range(1, 5)
            ]
            df = pd.DataFrame(
                tabela, columns=["Num. Subsistema", "Fictício"] + cols
            )
            df["Nome"] = subsistemas
            df = df[["Num. Subsistema", "Nome", "Fictício"] + cols]
            df = df.astype({"Num. Subsistema": "int64", "Fictício": "int64"})
            return df

        # Salta as linhas adicionais
        for _ in range(3):
            self.__cabecalhos.append(file.readline())

        i = 0
        subsistemas: List[str] = []
        tabela = np.zeros((MAX_SUBMERCADOS, 10))
        while True:
            linha = file.readline()
            # Confere se terminaram
            if len(linha) < 3 or BlocoCustosDeficit.FIM_BLOCO in linha:
                # Converte para df e salva na variável
                if i > 0:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                break
            # Confere se é uma linha de subsistema ou tabela
            else:
                dados = self.__linha.read(linha)
                tabela[i, 0] = dados[0]
                subsistemas.append(dados[1])
                tabela[i, 1:] = dados[2:]
                i += 1

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError(
                "Dados do sistema.dat não foram lidos com sucesso"
            )
        cols = [f"Custo Pat. {p}" for p in range(1, 5)] + [
            f"Corte Pat. {p}" for p in range(1, 5)
        ]
        for _, linha_dados in self.data.iterrows():
            dados_linha = linha_dados[cols].tolist()
            dados_linha_escrita = linha_dados.tolist()[:3]
            for d in dados_linha:
                dados_linha_escrita.append(d if not np.isnan(d) else None)
            file.write(self.__linha.write(dados_linha_escrita))
        file.write(BlocoCustosDeficit.FIM_BLOCO + "\n")


class BlocoIntercambioSubsistema(Section):
    """
    Bloco com a informação de intercâmbio
    por mês/ano de estudo para cada subsistema.
    """

    FIM_BLOCO = " 999"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha_subsis = Line(
            [IntegerField(3, 1), IntegerField(3, 5), IntegerField(1, 23)]
        )
        campo_ano: List[Field] = [IntegerField(4, 0)]
        campos_interc: List[Field] = [
            FloatField(6, 8 + i * 8, 0) for i in range(len(MESES_DF))
        ]
        self.__linha = Line(campo_ano + campos_interc)
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoIntercambioSubsistema):
            return False
        bloco: BlocoIntercambioSubsistema = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            df = pd.DataFrame(
                tabela,
                columns=["Ano"] + MESES_DF,
            )
            df["Subsistema De"] = subsistemas_de
            df["Subsistema Para"] = subsistemas_para
            df["Sentido"] = sentidos
            df = df[
                ["Subsistema De", "Subsistema Para", "Sentido", "Ano"]
                + MESES_DF
            ]
            df = df.astype(
                {
                    "Subsistema De": "int64",
                    "Subsistema Para": "int64",
                    "Sentido": "int64",
                    "Ano": "int64",
                }
            )
            return df

        # Salta as linhas adicionais
        for _ in range(3):
            self.__cabecalhos.append(file.readline())

        i = 0
        subsis_de_atual = 0
        subsis_para_atual = 0
        sentido_atual = -1
        ano_atual = 0
        subsistemas_de: List[int] = []
        subsistemas_para: List[int] = []
        sentidos: List[int] = []
        tabela = np.zeros(
            (
                MAX_SUBMERCADOS * MAX_SUBMERCADOS * MAX_ANOS_ESTUDO,
                len(MESES_DF) + 1,
            )
        )
        while True:
            linha = file.readline()
            # Confere se terminaram
            if BlocoIntercambioSubsistema.FIM_BLOCO in linha[:4] or (
                len(linha) < 3 and i == 0
            ):
                # Converte para df e salva na variável
                if i > 0:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                break
            # Confere se é uma linha de subsistema ou tabela
            if len(linha) < 30:
                dados = self.__linha_subsis.read(linha)
                subsis_de_atual = (
                    subsis_de_atual if dados[0] is None else dados[0]
                )
                subsis_para_atual = (
                    subsis_para_atual if dados[1] is None else dados[1]
                )
                sentido_atual = (
                    dados[2]
                    if dados[2] is not None
                    else int(not sentido_atual)
                )
            else:
                dados = self.__linha.read(linha)
                if isinstance(dados[0], int) and dados[0] != ano_atual:
                    ano_atual = dados[0]
                subsistemas_de.append(subsis_de_atual)
                subsistemas_para.append(subsis_para_atual)
                sentidos.append(sentido_atual)
                tabela[i, 0] = ano_atual
                tabela[i, 1:] = dados[1:]
                i += 1

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError(
                "Dados do sistema.dat não foram lidos com sucesso"
            )

        ultimo_subsistema_de = 0
        ultimo_subsistema_para = 0
        ultimo_sentido = -1

        for _, linha in self.data.iterrows():
            linha_lida: pd.Series = linha
            if any(
                [
                    linha_lida["Subsistema De"] != ultimo_subsistema_de,
                    linha_lida["Subsistema Para"] != ultimo_subsistema_para,
                    linha_lida["Sentido"] != ultimo_sentido,
                ]
            ):
                ultimo_ano = 0
                ultimo_subsistema_de = linha_lida["Subsistema De"]
                ultimo_subsistema_para = linha_lida["Subsistema Para"]
                ultimo_sentido = linha_lida["Sentido"]
                file.write(
                    self.__linha_subsis.write(
                        [
                            int(ultimo_subsistema_de),
                            int(ultimo_subsistema_para),
                            int(ultimo_sentido),
                        ]
                    )
                )
            ano_linha = (
                int(linha_lida["Ano"])
                if linha_lida["Ano"] != ultimo_ano
                else None
            )
            ultimo_ano = int(linha_lida["Ano"])
            dados_linha = linha_lida[MESES_DF].tolist()
            dados_linha_escrita = []
            for d in dados_linha:
                dados_linha_escrita.append(d if not np.isnan(d) else None)
            file.write(self.__linha.write([ano_linha] + dados_linha_escrita))
        file.write(BlocoIntercambioSubsistema.FIM_BLOCO + "\n")


class BlocoMercadoEnergiaSistema(Section):
    """
    Bloco com a informação de mercado de energia
    por mês/ano de estudo para cada subsistema.
    """

    FIM_BLOCO = " 999"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha_subsis = Line([IntegerField(3, 1)])
        campo_ano: List[Field] = [LiteralField(4, 0)]
        campos_mercado: List[Field] = [
            FloatField(7, 7 + i * 8, 0) for i in range(len(MESES_DF))
        ]
        self.__linha = Line(campo_ano + campos_mercado)
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoMercadoEnergiaSistema):
            return False
        bloco: BlocoMercadoEnergiaSistema = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            df = pd.DataFrame(
                tabela,
                columns=["Subsistema"] + MESES_DF,
            )
            df = df.astype(
                {
                    "Subsistema": "int64",
                }
            )
            df["Ano"] = anos
            df = df[["Subsistema", "Ano"] + MESES_DF]
            return df

        # Salta as linhas adicionais
        for _ in range(3):
            self.__cabecalhos.append(file.readline())

        i = 0
        subsis_atual = 0
        ano_atual = ""
        anos: List[str] = []
        tabela = np.zeros(
            (
                MAX_SUBMERCADOS * MAX_ANOS_ESTUDO,
                len(MESES_DF) + 1,
            )
        )
        while True:
            linha = file.readline()
            # Confere se terminaram
            if (
                len(linha) < 3
                or BlocoMercadoEnergiaSistema.FIM_BLOCO in linha[:4]
            ):
                # Converte para df e salva na variável
                if i > 0:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                break
            # Confere se é uma linha de subsistema ou tabela
            if len(linha) < 12:
                dados = self.__linha_subsis.read(linha)
                subsis_atual = dados[0]
            else:
                dados = self.__linha.read(linha)
                if isinstance(dados[0], str) and dados[0] != ano_atual:
                    ano_atual = dados[0]
                anos.append(ano_atual)
                tabela[i, 0] = subsis_atual
                tabela[i, 1:] = dados[1:]
                i += 1

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError(
                "Dados do sistema.dat não foram lidos com sucesso"
            )

        ultimo_subsistema = 0

        for _, linha in self.data.iterrows():
            linha_lida: pd.Series = linha
            if any(
                [
                    linha_lida["Subsistema"] != ultimo_subsistema,
                ]
            ):
                ultimo_ano = ""
                ultimo_subsistema = linha_lida["Subsistema"]
                file.write(
                    self.__linha_subsis.write(
                        [
                            int(ultimo_subsistema),
                        ]
                    )
                )
            ano_linha = (
                linha_lida["Ano"] if linha_lida["Ano"] != ultimo_ano else None
            )
            ultimo_ano = linha_lida["Ano"]
            dados_linha = linha_lida[MESES_DF].tolist()
            dados_linha_escrita = []
            for d in dados_linha:
                dados_linha_escrita.append(d if not np.isnan(d) else None)
            file.write(self.__linha.write([ano_linha] + dados_linha_escrita))
        file.write(BlocoMercadoEnergiaSistema.FIM_BLOCO + "\n")


class BlocoGeracaoUsinasNaoSimuladas(Section):
    """
    Bloco com a geração de usinas não simuladas em P.U. para
    cada patamar, por mês de estudo, extraído do arquivo `sistema.dat`.
    """

    FIM_BLOCO = " 999"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha_subsis = Line(
            [IntegerField(3, 1), IntegerField(3, 6), LiteralField(3, 11)]
        )

        campo_ano: List[Field] = [IntegerField(4, 0)]
        campos_geradores: List[Field] = [
            FloatField(7, 7 + i * 8, 0) for i in range(len(MESES_DF))
        ]
        self.__linha = Line(campo_ano + campos_geradores)
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoGeracaoUsinasNaoSimuladas):
            return False
        bloco: BlocoGeracaoUsinasNaoSimuladas = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            df = pd.DataFrame(
                tabela, columns=["Subsistema", "Bloco", "Ano"] + MESES_DF
            )
            df["Razão"] = razoes
            df = df.astype({"Subsistema": "int64", "Ano": "int64"})
            df = df[["Subsistema", "Bloco", "Razão", "Ano"] + MESES_DF]
            return df

        # Salta as linhas adicionais
        for _ in range(3):
            self.__cabecalhos.append(file.readline())

        i = 0
        subsis_atual = 0
        bloco_atual = 0
        razao_atual = ""
        ano_atual = 0
        razoes: List[str] = []
        tabela = np.zeros(
            (
                MAX_SUBMERCADOS * MAX_SUBMERCADOS * MAX_ANOS_ESTUDO,
                len(MESES_DF) + 3,
            )
        )
        while True:
            linha = file.readline()
            # Confere se terminaram
            if (
                BlocoGeracaoUsinasNaoSimuladas.FIM_BLOCO in linha[:4]
                or len(linha) < 3
            ):
                # Converte para df e salva na variável
                if i > 0:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                break
            # Confere se é uma linha de subsistema ou tabela
            if len(linha) < 30:
                dados = self.__linha_subsis.read(linha)
                subsis_atual = dados[0]
                bloco_atual = dados[1]
                razao_atual = dados[2]
            else:
                dados = self.__linha.read(linha)
                if isinstance(dados[0], int) and dados[0] != ano_atual:
                    ano_atual = dados[0]
                razoes.append(razao_atual)
                tabela[i, 0] = subsis_atual
                tabela[i, 1] = bloco_atual
                tabela[i, 2] = ano_atual
                tabela[i, 3:] = dados[1:]
                i += 1

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError(
                "Dados do sistema.dat não foram lidos com sucesso"
            )

        ultimo_subsistema: Optional[int] = 0
        ultimo_bloco: Optional[int] = 0
        ultima_razao: Optional[str] = ""

        for _, linha in self.data.iterrows():
            linha_lida: pd.Series = linha
            novo_subsis = (
                int(linha_lida["Subsistema"])
                if not np.isnan(linha_lida["Subsistema"])
                else None
            )
            novo_bloco = (
                int(linha_lida["Bloco"])
                if not np.isnan(linha_lida["Bloco"])
                else None
            )
            nova_razao = (
                str(linha_lida["Razão"])
                if not linha_lida["Razão"] is None
                else None
            )
            if any(
                [
                    novo_subsis != ultimo_subsistema,
                    novo_bloco != ultimo_bloco,
                    nova_razao != ultima_razao,
                ]
            ):
                ultimo_ano = 0
                ultimo_subsistema = novo_subsis
                ultimo_bloco = novo_bloco
                ultima_razao = nova_razao
                file.write(
                    self.__linha_subsis.write(
                        [
                            novo_subsis,
                            novo_bloco,
                            nova_razao,
                        ]
                    )
                )
            ano_linha = (
                int(linha_lida["Ano"])
                if linha_lida["Ano"] != ultimo_ano
                else None
            )
            ultimo_ano = int(linha_lida["Ano"])
            dados_linha = linha_lida[MESES_DF].tolist()
            dados_linha_escrita = []
            for d in dados_linha:
                dados_linha_escrita.append(d if not np.isnan(d) else None)
            file.write(self.__linha.write([ano_linha] + dados_linha_escrita))
        file.write(BlocoGeracaoUsinasNaoSimuladas.FIM_BLOCO + "\n")



from inewave.config import MAX_REES

from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.integerfield import IntegerField
from typing import List, IO
import pandas as pd  # type: ignore
import numpy as np  # type: ignore


class BlocoReesSubmercados(Section):
    """
    Bloco com informações dos REEs e suas relações com os
    submercados.
    """

    FIM_BLOCO = "999"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line(
            [
                IntegerField(3, 1),
                LiteralField(10, 5),
                IntegerField(3, 18),
                IntegerField(2, 23),
                IntegerField(4, 26),
            ]
        )
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoReesSubmercados):
            return False
        bloco: BlocoReesSubmercados = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            cols = [
                "Número",
                "Submercado",
                "Mês Fim Individualizado",
                "Ano Fim Individualizado",
            ]
            df = pd.DataFrame(tabela, columns=cols)
            df["Nome"] = nomes
            df = df.astype({"Submercado": np.int64, "Número": np.int64})
            df = df[
                [
                    "Número",
                    "Nome",
                    "Submercado",
                    "Mês Fim Individualizado",
                    "Ano Fim Individualizado",
                ]
            ]
            return df

        # Salta as linhas adicionais
        for _ in range(3):
            self.__cabecalhos.append(file.readline())

        i = 0
        nomes: List[str] = []
        tabela = np.zeros((MAX_REES, 4))
        while True:
            linha = file.readline()
            # Confere se terminaram
            if len(linha) < 3 or BlocoReesSubmercados.FIM_BLOCO in linha:
                # Converte para df e salva na variável
                if i > 0:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                break
            # Confere se é uma linha de subsistema ou tabela
            else:
                dados = self.__linha.read(linha)
                tabela[i, :] = [dados[0]] + dados[2:]
                nomes.append(dados[1])
                i += 1

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError("Dados do ree.dat não foram lidos com sucesso")

        for _, lin in self.data.iterrows():
            linha_escrita = lin.tolist()
            for i, v in enumerate(linha_escrita[2:], start=2):
                linha_escrita[i] = None if np.isnan(v) else int(v)
            file.write(self.__linha.write(linha_escrita))
        file.write(BlocoReesSubmercados.FIM_BLOCO + "\n")


class BlocoFicticiasIndividualizado(Section):
    """
    Bloco com a opção de remover usinas fictícias nos
    períodos individualizados.
    """

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line([LiteralField(20, 0), IntegerField(4, 21)])

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoFicticiasIndividualizado):
            return False
        bloco: BlocoFicticiasIndividualizado = o
        if not all(
            [
                isinstance(self.data, list),
                isinstance(o.data, list),
            ]
        ):
            return False
        else:
            return self.data == bloco.data

    def read(self, file: IO):
        self.data = self.__linha.read(file.readline())

    def write(self, file: IO):
        file.write(self.__linha.write(self.data))



from inewave.nwlistcf.modelos.estados import EstadosPeriodoNwlistcf

from cfinterface.files.blockfile import BlockFile
from typing import TypeVar, Optional
import pandas as pd  # type: ignore


class Estados(BlockFile):
    """
    Armazena os dados dos estados visitados pelo NEWAVE existentes
    no arquivo `estados.rel` do NWLISTCF.

    Esta classe armazena os estados de cada uma das variáveis envolvidas
    no problema e da função objetivo, para cada registro e REE dentro
    do registro.

    """

    T = TypeVar("T")

    BLOCKS = [EstadosPeriodoNwlistcf]

    def __init__(self, data=...) -> None:
        super().__init__(data)
        self.__estados_periodos = None

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="estados.rel"
    ) -> "Estados":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="estados.rel"):
        self.write(diretorio, nome_arquivo)

    def __monta_tabela_estados(self) -> pd.DataFrame:
        df = None
        for b in self.data.of_type(EstadosPeriodoNwlistcf):
            dados = b.data
            if dados is None:
                continue
            elif df is None:
                df = b.data
            else:
                df = pd.concat([df, b.data], ignore_index=True)
        return df

    @property
    def estados(self) -> Optional[pd.DataFrame]:
        """
        Tabela com os estados visitados na construção da FCF.

        - PERIODO (`int`)
        - IREG (`int`)
        - ITEc (`int`)
        - SIMc (`int`)
        - ITEf (`int`)
        - REE (`int`)
        - FUNC. OBJ. (`float`)
        - EARM (`float`)
        - EAF 1 (`float`)
        - ...
        - EAF 6 (`float`)
        - SGT(P1E1) (`float`)
        - ...
        - SGT(P3E2) (`float`)
        - MX_SAR (`float`)
        - MX_CURVA (`float`)

        :return: A tabela de estados como um DataFrame
        :rtype: pd.DataFrame | None
        """
        if self.__estados_periodos is None:
            self.__estados_periodos = self.__monta_tabela_estados()
        return self.__estados_periodos
